using Phares.Shared.Models;
using Phares.Shared.Models.Properties;
using Phares.Shared.Models.Stateless;
using System.Collections.ObjectModel;

namespace View_by_Distance.Distance.Models.Stateless;

internal static class FaceEncodingLogic
{

    internal static void MoveUnableToMatch(FilePath filePath)
    {
        string checkFile = $"{filePath.FullName}.unk";
        if (File.Exists(filePath.FullName) && !File.Exists(checkFile))
            File.Move(filePath.FullName, checkFile);
    }

    internal static ReadOnlyDictionary<int, ReadOnlyDictionary<int, LocationContainer>> GetMappedIdsThenWholePercentagesToLocationContainer(ResultSettings resultSettings, MetadataSettings metadataSettings, PeopleSettings peopleSettings, DistanceSettings distanceSettings, ICompareSettings compareSettings, ICompare compare, ReadOnlyCollections readOnlyCollections)
    {
        ReadOnlyDictionary<int, ReadOnlyDictionary<int, LocationContainer>> results;
        List<LocationContainer?> locationContainers = [];
        int maxDegreeOfParallelism = compareSettings.MaxDegreeOfParallelism;
        ParallelOptions parallelOptions = new() { MaxDegreeOfParallelism = maxDegreeOfParallelism };
        ReadOnlyCollection<ExifDirectory> exifDirectories = MappedLogicA.GetMapped(resultSettings,
                                                                                   metadataSettings,
                                                                                   peopleSettings,
                                                                                   distanceSettings,
                                                                                   compareSettings,
                                                                                   compare,
                                                                                   readOnlyCollections);
        int totalSeconds = (int)Math.Floor(new TimeSpan(DateTime.Now.Ticks - compare.Ticks).TotalSeconds);
        string message = $") Building Mapped with Encoding Face Files Collection - {totalSeconds} total second(s)";
        compare.ConstructProgressBar(exifDirectories.Count, message);
        _ = Parallel.For(0, exifDirectories.Count, parallelOptions, (i, state) =>
            LocationContainersParallelFor(distanceSettings, compareSettings, compare, exifDirectories, i, locationContainers));
        results = GetMappedIdsThenWholePercentagesToLocationContainer(locationContainers);
        return results;
    }

    private static void LocationContainersParallelFor(DistanceSettings distanceSettings, ICompareSettings compareSettings, ICompare compare, ReadOnlyCollection<ExifDirectory> exifDirectories, int i, List<LocationContainer?> results)
    {
        compare.Tick();
        ExifDirectory exifDirectory = exifDirectories[i];
        LocationContainer? locationContainer = GetLocationContainer(distanceSettings, compareSettings, exifDirectory);
        results.Add(locationContainer);
    }

    internal static LocationContainer? GetLocationContainer(DistanceSettings distanceSettings, ICompareSettings compareSettings, ExifDirectory exifDirectory)
    {
        LocationContainer? result;
        DateOnly dateOnly;
        FaceFile? faceFile;
        int? wholePercentages;
        FaceEncoding? faceEncoding;
        FaceRecognitionDotNet.Models.FaceEncoding? encoding;
        wholePercentages = IMapping.GetWholePercentages(compareSettings, exifDirectory.FilePath);
        if (wholePercentages is null)
        {
            if (distanceSettings.DistanceMoveUnableToMatch)
                MoveUnableToMatch(exifDirectory.FilePath);
            result = null;
        }
        else
        {
            faceFile = GetFaceFile(distanceSettings, exifDirectory);
            if (faceFile is null)
                result = null;
            else
            {
                dateOnly = DateOnly.FromDateTime(new DateTime(exifDirectory.FilePath.CreationTicks));
                faceEncoding = Phares.Metadata.Models.Stateless.IMetadata.GetFaceEncoding(exifDirectory);
                if (faceEncoding is null)
                    result = null;
                else
                {
                    encoding = FaceRecognitionDotNet.Models.FaceRecognition.LoadFaceEncoding(faceEncoding.RawEncoding);
                    result = new(CreationDateOnly: dateOnly,
                                 ExifDirectory: exifDirectory,
                                 Encoding: encoding,
                                 FaceFile: faceFile,
                                 FilePath: exifDirectory.FilePath,
                                 LengthPermyriad: null,
                                 LengthSource: null,
                                 PersonKeyFormattedAndKeyTicksAndDisplayDirectoryName: exifDirectory.PersonKeyFormattedAndKeyTicksAndDisplayDirectoryName,
                                 WholePercentages: wholePercentages);
                }
            }
        }
        return result;
    }

    private static FaceFile? GetFaceFile(DistanceSettings distanceSettings, ExifDirectory exifDirectory)
    {
        FaceFile? result = Phares.Metadata.Models.Stateless.IMetadata.GetFaceFile(exifDirectory);
        if (result is null || result.Location is null)
        {
            if (distanceSettings.DistanceMoveUnableToMatch)
                MoveUnableToMatch(exifDirectory.FilePath);
            result = null;
        }
        return result;
    }

    private static ReadOnlyDictionary<int, ReadOnlyDictionary<int, LocationContainer>> GetMappedIdsThenWholePercentagesToLocationContainer(List<LocationContainer?> locationContainers)
    {
        Dictionary<int, ReadOnlyDictionary<int, LocationContainer>> results = [];
        int id;
        int wholePercentages;
        Dictionary<int, LocationContainer>? keyValue;
        Dictionary<int, Dictionary<int, LocationContainer>> keyValuePairs = [];
        foreach (LocationContainer? locationContainer in locationContainers)
        {
            if (locationContainer?.FilePath.Id is null || locationContainer.WholePercentages is null)
                continue;
            id = locationContainer.FilePath.Id.Value;
            wholePercentages = locationContainer.WholePercentages.Value;
            if (!keyValuePairs.TryGetValue(id, out keyValue))
            {
                keyValuePairs.Add(id, []);
                if (!keyValuePairs.TryGetValue(id, out keyValue))
                    throw new Exception();
            }
            keyValue.Add(wholePercentages, locationContainer);
        }
        foreach (KeyValuePair<int, Dictionary<int, LocationContainer>> keyValuePair in keyValuePairs)
            results.Add(keyValuePair.Key, keyValuePair.Value.AsReadOnly());
        return results.AsReadOnly();
    }

}